#include <SPI.h>
#include <nRF24L01.h>
#include <RF24.h>

RF24 radio(3, 4);  // CE, CSN
const byte address[6] = "ABCDE";

struct packetdata
{
  byte pot1;
  byte pot2;
  byte pot3;
  byte swtch1;
};
packetdata receiverdata;

// Uncomment to print the MPU data on the serial monitor for debugging
//#define PRINT_DEBUG_BUILD

// PID library
#include <PID_v1.h>

// These are needed for MPU
#include "I2Cdev.h"
#include "MPU6050_6Axis_MotionApps20.h"

#if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
    #include "Wire.h"
#endif

// class default I2C address is 0x68
MPU6050 mpu;

#define INTERRUPT_PIN 2  // use pin 2 on Arduino Uno & most boards

// MPU control/status vars
bool dmpReady = false;  // set true if DMP init was successful
uint8_t mpuIntStatus;   // holds actual interrupt status byte from MPU
uint8_t devStatus;      // return status after each device operation (0 = success, !0 = error)
uint16_t packetSize;    // expected DMP packet size (default is 42 bytes)
uint16_t fifoCount;     // count of all bytes currently in FIFO
uint8_t fifoBuffer[64]; // FIFO storage buffer

// orientation/motion vars
Quaternion q;           // [w, x, y, z] quaternion container
VectorFloat gravity;    // [x, y, z] gravity vector
float ypr[3];           // [yaw, pitch, roll] yaw/pitch/roll container and gravity vector
VectorInt16 gy;         // [x, y, z] gyro sensor measurements

// ================================================================
// === INTERRUPT DETECTION ROUTINE ===
// ================================================================

volatile bool mpuInterrupt = false;  // indicates whether MPU interrupt pin has gone high
void dmpDataReady() {
    mpuInterrupt = true;
}

#define PID_MIN_LIMIT -255  // Min limit for PID 
#define PID_MAX_LIMIT 255   // Max limit for PID 
#define PID_SAMPLE_TIME_IN_MILLI 10  // PID sample time in milliseconds

#define SETPOINT_PITCH_ANGLE_OFFSET 0  // Angle correction to keep robot standing vertically
#define MIN_ABSOLUTE_SPEED 15  // Min motor speed 

double setpointPitchAngle = SETPOINT_PITCH_ANGLE_OFFSET;
double pitchGyroAngle = 0;
double pitchPIDOutput = 0;

double setpointYawRate = 0;
double yawGyroRate = 0;
double yawPIDOutput = 0;

#define PID_PITCH_KP 68
#define PID_PITCH_KI 600
#define PID_PITCH_KD 1.5

#define PID_YAW_KP 0.5
#define PID_YAW_KI 0.3 
#define PID_YAW_KD 0.01

PID pitchPID(&pitchGyroAngle, &pitchPIDOutput, &setpointPitchAngle, PID_PITCH_KP, PID_PITCH_KI, PID_PITCH_KD, DIRECT);
PID yawPID(&yawGyroRate, &yawPIDOutput, &setpointYawRate, PID_YAW_KP, PID_YAW_KI, PID_YAW_KD, DIRECT);

int enableMotor2 = 10;
int motor2Pin1 = 7;
int motor2Pin2 = 8;

int motor1Pin1 = 6;
int motor1Pin2 = 5;
int enableMotor1 = 9;

void setupPID() {
    pitchPID.SetOutputLimits(PID_MIN_LIMIT, PID_MAX_LIMIT);
    pitchPID.SetMode(AUTOMATIC);
    pitchPID.SetSampleTime(PID_SAMPLE_TIME_IN_MILLI);

    yawPID.SetOutputLimits(PID_MIN_LIMIT, PID_MAX_LIMIT);
    yawPID.SetMode(AUTOMATIC);
    yawPID.SetSampleTime(PID_SAMPLE_TIME_IN_MILLI);
}

void setupMotors() {
    pinMode(enableMotor1, OUTPUT);
    pinMode(motor1Pin1, OUTPUT);
    pinMode(motor1Pin2, OUTPUT);
  
    pinMode(enableMotor2, OUTPUT);
    pinMode(motor2Pin1, OUTPUT);
    pinMode(motor2Pin2, OUTPUT);

    pinMode(A0, OUTPUT);
    rotateMotor(0, 0);  // Initialize motors to stop
}

void setupMPU() {
    #if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
        Wire.begin();
        Wire.setClock(400000); // 400kHz I2C clock. Comment this line if having compilation difficulties
    #elif I2CDEV_IMPLEMENTATION == I2CDEV_BUILTIN_FASTWIRE
        Fastwire::setup(400, true);
    #endif
  
    mpu.initialize();
    pinMode(INTERRUPT_PIN, INPUT);
    devStatus = mpu.dmpInitialize();

    // Use predefined offsets instead of auto-calibration
    mpu.setXAccelOffset(-429); 
    mpu.setYAccelOffset(-1701); 
    mpu.setZAccelOffset(1161);   
    mpu.setXGyroOffset(-5);
    mpu.setYGyroOffset(-6);
    mpu.setZGyroOffset(15);

    if (devStatus == 0) {
        mpu.setDMPEnabled(true);
        mpuIntStatus = mpu.getIntStatus();
        dmpReady = true;
        packetSize = mpu.dmpGetFIFOPacketSize();
    } else {
        Serial.print(F("DMP Initialization failed (code "));
        Serial.print(devStatus);
        Serial.println(F(")"));
    }
}

void setup() {
    Serial.begin(115200);
    radio.begin();
    radio.openReadingPipe(0, address);
    radio.setPALevel(RF24_PA_MIN);
    radio.setDataRate(RF24_250KBPS);
    radio.setChannel(124);
    radio.startListening();
    setupMotors();   
    setupMPU();
    setupPID();
}

void loop() {
    if (radio.available()) {
        radio.read(&receiverdata, sizeof(packetdata));
        char inputvaluestring[100];
        sprintf(inputvaluestring,"%3d,%3d,%3d,%3d",receiverdata.pot1,receiverdata.pot2,receiverdata.pot3,receiverdata.swtch1);
        Serial.println(inputvaluestring);
        digitalWrite(A0, receiverdata.swtch1);
    } 
    
    if (!dmpReady) return;

    if (mpu.dmpGetCurrentFIFOPacket(fifoBuffer)) {  
        mpu.dmpGetQuaternion(&q, fifoBuffer);
        mpu.dmpGetGravity(&gravity, &q);
        mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
        mpu.dmpGetGyro(&gy, fifoBuffer);

        pitchGyroAngle = ypr[1] * 180/M_PI;  // angle in degrees
        yawGyroRate = gy.z;  // rotation rate in degrees per second

        pitchPID.Compute();
        yawPID.Compute();

        rotateMotor(pitchPIDOutput + yawPIDOutput, pitchPIDOutput - yawPIDOutput);
        
        #ifdef PRINT_DEBUG_BUILD
            Serial.print("Yaw Rate: ");
            Serial.println(yawGyroRate);
            Serial.print("Yaw PID Output: ");
            Serial.println(yawPIDOutput);
            Serial.println(setpointPitchAngle);
            Serial.println("The PID output ");
            Serial.println(pitchPIDOutput);
            delay(500);   
        #endif
    }
}

void rotateMotor(int speed1, int speed2) {
    if (speed1 < 0) {
        digitalWrite(motor1Pin1, HIGH);
        digitalWrite(motor1Pin2, LOW);    
    } else {
        digitalWrite(motor1Pin1, LOW);
        digitalWrite(motor1Pin2, HIGH);      
    }

    if (speed2 < 0) {
        digitalWrite(motor2Pin1, HIGH);
        digitalWrite(motor2Pin2, LOW);    
    } else {
        digitalWrite(motor2Pin1, LOW);
        digitalWrite(motor2Pin2, HIGH);      
    }

    speed1 = abs(speed1) + MIN_ABSOLUTE_SPEED;
    speed2 = abs(speed2) + MIN_ABSOLUTE_SPEED;

    speed1 = constrain(speed1, MIN_ABSOLUTE_SPEED, 255);
    speed2 = constrain(speed2, MIN_ABSOLUTE_SPEED, 255);

    analogWrite(enableMotor1, speed1);
    analogWrite(enableMotor2, speed2);
}
